###########################################################################
#
# OpenOPC for Python OPC-DA Library File
#
# A Windows only system metrics library file.
#
# Copyright (c) 2007-2012 Barry Barnreiter (barry_b@users.sourceforge.net)
# Copyright (c) 2014 Anton D. Kachalov (mouse@yandex.ru)
# Copyright (c) 2017 José A. Maita (jose.a.maita@gmail.com)
# Copyright (c) 2022 j3mg
#
###########################################################################
import win32com.client
import win32process
import win32pdh
import win32pdhquery
import win32pdhutil
import pywintypes
import ctypes
import time
import os
import math

# NT4 does not have WMI by default
try:
    import wmi
except:
    wmi_found = False
else:
    wmi_found = True

class CPU:
    def __init__(self):
        path = win32pdh.MakeCounterPath((None, "Processor", "_Total", None, -1, "% Processor Time"))
        self.base = win32pdh.OpenQuery()
        self.counter = win32pdh.AddCounter(self.base, path)
        self.reset()

    def reset(self):
        win32pdh.CollectQueryData(self.base)

    def get_usage(self):
        win32pdh.CollectQueryData(self.base)
        try:
            value = win32pdh.GetFormattedCounterValue(self.counter, win32pdh.PDH_FMT_LONG)[1]
        except pywintypes.error:
            value = 0
        return value

def _disk_info():
    drive = os.getenv("SystemDrive")
    freeuser = ctypes.c_int64()
    total = ctypes.c_int64()
    free = ctypes.c_int64()
    ctypes.windll.kernel32.GetDiskFreeSpaceExW(drive, ctypes.byref(freeuser), ctypes.byref(total), ctypes.byref(free))
    return freeuser.value 

def disk_free():
    return int(_disk_info() / 1024)

def _mem_info():
        kernel32 = ctypes.windll.kernel32
        c_ulong = ctypes.c_ulong
        class MEMORYSTATUS(ctypes.Structure):
            _fields_ = [
                ('dwLength', c_ulong),
                ('dwMemoryLoad', c_ulong),
                ('dwTotalPhys', c_ulong),
                ('dwAvailPhys', c_ulong),
                ('dwTotalPageFile', c_ulong),
                ('dwAvailPageFile', c_ulong),
                ('dwTotalVirtual', c_ulong),
                ('dwAvailVirtual', c_ulong)
            ]
            
        memoryStatus = MEMORYSTATUS()
        memoryStatus.dwLength = ctypes.sizeof(MEMORYSTATUS)
        kernel32.GlobalMemoryStatus(ctypes.byref(memoryStatus))
        return (memoryStatus.dwTotalPhys, memoryStatus.dwAvailPhys)

def mem_used():
    counter=r'\Memory\Committed Bytes'
    machine, object, instance, parentInstance, index, counter = win32pdh.ParseCounterPath(counter)

    instance = None
    inum=-1
    format = win32pdh.PDH_FMT_DOUBLE
    machine=None
    
    path = win32pdh.MakeCounterPath( (machine,object, instance, None, inum,counter) )
    hq = win32pdh.OpenQuery()
    try:
        hc = win32pdh.AddCounter(hq, path)
        try:
            win32pdh.CollectQueryData(hq)
            type, val = win32pdh.GetFormattedCounterValue(hc, format)
            return int(val / 1024)
        except pywintypes.error:
            return 0
        finally:
            win32pdh.RemoveCounter(hc)
    finally:
        win32pdh.CloseQuery(hq)

def mem_free():
    total,free = _mem_info()
    return int(free / 1024)

def mem_total():
    total,free = _mem_info()
    return int(total / 1024)

def mem_percent():
    total,free = _mem_info()
    return ( float(total - free) / float(total) ) * 100.0
    
def _task_list():
    psapi = ctypes.windll.psapi
    kernel = ctypes.windll.kernel32

    hModule = ctypes.c_ulong()
    count = ctypes.c_ulong()
    modname = ctypes.c_buffer(30)
    PROCESS_QUERY_INFORMATION = 0x0400
    PROCESS_VM_READ = 0x0010

    pid_list = win32process.EnumProcesses()
    info_list = []
    
    for pid in pid_list:
        
        hProcess = kernel.OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, False, pid)
        if hProcess:
            psapi.EnumProcessModules(hProcess, ctypes.byref(hModule), ctypes.sizeof(hModule), ctypes.byref(count))
            psapi.GetModuleBaseNameA(hProcess, hModule.value, modname, ctypes.sizeof(modname))
            pname = ctypes.string_at(modname)

            procmeminfo = win32process.GetProcessMemoryInfo(hProcess)
            procmemusage = (procmeminfo["WorkingSetSize"]/1024)
            info_list.append((pid, pname, procmemusage))
                        
            kernel.CloseHandle(hProcess)

    return info_list

def task_mem(image_name):
    image_name = str.lower(image_name)
    if image_name[-4:] != '.exe': image_name = image_name + '.exe'
    return sum([mem for pid,name,mem in _task_list() if str.lower(str(name)) == image_name])

def task_exists(image_name):
    image_name = str.lower(image_name)
    if image_name[-4:] != '.exe': image_name = image_name + '.exe'
    return len([mem for pid,name,mem in _task_list() if str.lower(str(name)) == image_name]) > 0

def task_cpu(image_name):
    if not wmi_found: return 0.0

    image_name = str.lower(image_name)
    if image_name[-4:] == '.exe': image_name = image_name[:-4]

    c = wmi.WMI()
    process_info = {}
    pct_cpu_time = 0.0

    for i in range(2):
    
        for p in c.Win32_PerfRawData_PerfProc_Process(name=image_name):
            id = long(p.IDProcess)
            n1, d1 = long(p.PercentProcessorTime), long(p.Timestamp_Sys100NS)
            n0, d0, so_far = process_info.get(id, (0, 0, []))

            try:
                pct_cpu_time += (float (n1 - n0) / float (d1 - d0)) * 100.0
            except ZeroDivisionError:
                pct_cpu_time += 0.0
            
            so_far.append(pct_cpu_time)
            process_info[id] = (n1, d1, so_far)

        if i == 0:
            time.sleep(0.1)
            pct_cpu_time = 0.0

    num_cpu = int(os.environ['NUMBER_OF_PROCESSORS'])
    return min(pct_cpu_time / num_cpu, 100.0)

def sine_wave():
    min = float(time.localtime()[4])
    sec = float(time.localtime()[5])
    T = (min + (sec/60.0)) % 10.0
    return math.sin(2.0 * math.pi * T/10.0) * 100.0

def saw_wave():
    min = float(time.localtime()[4])
    sec = float(time.localtime()[5])
    T = (min + (sec/60.0)) % 10.0
    return (T/10.0) * 100.0
    



###########################################################################
#
# OpenOPC Command Line Client
#
# A cross-platform OPC-DA client built using the OpenOPC for Python
# library module.
#
# Copyright (c) 2007-2012 Barry Barnreiter (barry_b@users.sourceforge.net)
# Copyright (c) 2014 Anton D. Kachalov (mouse@yandex.ru)
#
###########################################################################
from sys import *
from getopt import *
from os import *
import signal
import sys
import os
import types
import datetime
import re, time, csv
import OpenOPC

try:
    import Pyro5
except ImportError:
    pyro_found = False
else:
    pyro_found = True

# FUNCTION: Print comand line usage summary

def usage():
    print('OpenOPC Command Line Client', OpenOPC.__version__)
    print('Copyright (c) 2007-2012 Barry Barnreiter (barry_b@users.sourceforge.net)')
    print('')
    print('Usage:  opc [OPTIONS] [ACTION] [ITEM|PATH...]')
    print('')
    print('Actions:')
    print('  -r, --read                 Read ITEM values (default action)')
    print('  -w, --write                Write values to ITEMs (use ITEM=VALUE)')
    print('  -p, --properties           View properties of ITEMs')
    print('  -l, --list                 List items at specified PATHs (tree browser)')
    print('  -f, --flat                 List all ITEM names (flat browser)')
    print('  -i, --info                 Display OPC server information')
    print('  -q, --servers              Query list of available OPC servers')
    print('  -S, --sessions             List sessions in OpenOPC Gateway Service')
    print('')
    print('Options:')
    print('  -m MODE, --mode=MODE       Protocol MODE (dcom, open) (default: OPC_MODE)')
    print('  -C CLASS,--class=CLASS     OPC Automation CLASS (default: OPC_CLASS)')
    print('  -n NAME, --name=NAME       Set OPC Client NAME (default: OPC_CLIENT)')
    print('  -h HOST, --host=HOST       DCOM OPC HOST (default: OPC_HOST)')
    print('  -s SERV, --server=SERVER   DCOM OPC SERVER (default: OPC_SERVER)')
    print('  -H HOST, --gate-host=HOST  OpenOPC Gateway HOST (default: OPC_GATE_HOST)')
    print('  -P PORT, --gate-port=PORT  OpenOPC Gateway PORT (default: OPC_GATE_PORT)')
    print('')
    print('  -F FUNC, --function=FUNC   Read FUNCTION to use (sync, async)')
    print('  -c SRC,  --source=SOURCE   Set data SOURCE for reads (cache, device, hybrid)')
    print('  -g SIZE, --size=SIZE       Group tags into SIZE items per transaction')
    print('  -z MSEC, --pause=MSEC      Sleep MSEC milliseconds between transactions')
    print('  -u MSEC, --update=MSEC     Set update rate for group to MSEC milliseconds')
    print('  -t MSEC, --timeout=MSEC    Set read timeout to MSEC mulliseconds')
    print('')
    print('  -o FMT,  --output=FORMAT   Output FORMAT (table, values, pairs, csv, html)')
    print('  -L SEC,  --repeat=SEC      Loop ACTION every SEC seconds until stopped')
    print('  -y ID,   --id=ID,...       Retrieve only specific Property IDs')
    print('  -a STR,  --append=STR,...  Append STRINGS to each input item name')
    print('  -x N     --rotate=N        Rotate output orientation in groups of N values')
    print('  -v,      --verbose         Verbose mode showing all OPC function calls')
    print('  -e,      --errors          Include descriptive error message strings')
    print('  -R,      --recursive       List items recursively when browsing tree')
    print('  -,       --pipe            Pipe item/value list from standard input')

# Helper class for handling signals (i.e. Ctrl-C)

class SigHandler:
    def __init__(self):
        self.signaled = 0
        self.sn = None
    def __call__(self, sn, sf):
        self.sn = sn 
        self.signaled += 1

# FUNCTION: Iterable version of rotate()

def irotate(data, num_columns, value_idx = 1):
    if num_columns == 0:
        for row in data: yield row
        return
   
    new_row = []

    for i, row in enumerate(data):
        if type(row) not in (list, tuple):
            value_idx = 0
            row = [row]

        new_row.append(row[value_idx])
        if (i + 1) % num_columns == 0:
            yield new_row
            new_row = []

        if len(new_row) > 0:
            yield new_row

# FUNCTION: Rotate the values of every N rows to form N columns

def rotatelist(data, num_columns, value_idx = 1):
    return list(irotate(data, num_columns, value_idx))

# FUNCTION: Print output in the specified style from a list of data
   
def output(data, style = 'table', value_idx = 1):
    write = sys.stdout.write
    #global write
    name_idx = 0

    # Cast value to a stirng (trap Unicode errors)
    def to_str(value):
        try:
            if type(value) == float:
                return '%.4f' % value
            else:
                return str(value)
        except:
            return ''

    # Generator passed (single row passed at a time)
    if type(data) == types.GeneratorType:
        generator = True
        pad_length = []

    # List passed (multiple rows passed all at once)
    elif type(data) in (list, tuple):
        generator = False

        if len(data) == 0: return

        if type(data[0]) not in (list, tuple):
            data = [[e] for e in data]
         
        if style == 'table' or style == '':
            pad_length = []
            num_columns = len(data[0])
            for i in range(num_columns-1):
                pad_length.append(len(max([to_str(row[i]) for row in data], key=len)) + 5)
            pad_length.append(0)
    else:
        raise TypeError("output(): 'data' parameter must be a list or a generator")

    if style == 'html':
        write('<table border=1>\n')

    rows = []

    for i, row in enumerate(data):
        rows.append(row)

        if style == 'values':
            write('%s' % str(row[value_idx]))
        elif style == 'pairs':
            write('%s,%s' % (row[name_idx], row[value_idx]))
        else:

            if generator and (style == 'table' or style == ''):

                # Convert single value into a single element list, thus making it
                # represent a 1-column wide table.
                if type(row) not in (list, tuple):
                    row = [row]

                num_columns = len(row)

                # Allow columns widths to always grow wider, but never shrink.
                # Unfortunetly we won't know the required width until the generator is finished!
                for k in range(num_columns-1):
                    new_length = len(to_str(row[k]))
                    if i == 0:
                        pad_length.append(new_length + 5)
                    else:
                        if new_length - pad_length[k] > 0:  pad_length[k] = new_length
                if i == 0:
                    pad_length.append(0)

            for j, item in enumerate(row):               
                if style == 'csv':
                    if j > 0: write(',')
                    write('%s' % to_str(item))
                elif style == 'html':
                    if j == 0: write('  <tr>\n')
                    if len(to_str(item)) < 40:
                        write('    <td nowrap>%s</td>\n' % to_str(item))
                    else:
                        write('    <td>%s</td>\n' % to_str(item))
                    if j == len(row)-1: write('  </tr>')
                else:
                    if num_columns > 1:
                        write('%s' % to_str(item).ljust(pad_length[j]))
                    else:
                        write('%s' % to_str(item))

        write('\n')

    if style == 'html':
        write('</table>')

    return rows

# FUNCTION: Convert Unix time to formatted time string

def time2str(t):
    d = datetime.datetime.fromtimestamp(t)
    return d.strftime('%x %H:%M:%S')


######## MAIN ########
def main():

    # Common function aliases

    #write = sys.stdout.write

    # Initialize default settings
    
    if os.name == 'nt':
       opc_mode = 'dcom'
    else:
       opc_mode = 'open'

    opc_class = OpenOPC.OPC_CLASS
    client_name = OpenOPC.OPC_CLIENT
    opc_host = 'localhost'
    opc_server = OpenOPC.OPC_SERVER
    open_host = 'localhost'
    open_port = 7766

    action = 'read'
    style = 'table'
    append = ''
    num_columns = 0
    pipe = False
    verbose = False
    recursive = False
    read_function = 'async'
    data_source = 'hybrid'
    group_size = None
    update_rate = None
    timeout = 5000
    tx_pause = 0
    repeat = 1
    repeat_pause = None
    property_ids = None
    include_err_msg = False

    if 'OPC_MODE' in os.environ:         opc_mode = environ['OPC_MODE']
    if 'OPC_CLASS' in os.environ:        opc_class = environ['OPC_CLASS']
    if 'OPC_CLIENT' in os.environ:       client_name = environ['OPC_CLIENT']
    if 'OPC_HOST' in os.environ:         opc_host = environ['OPC_HOST']
    if 'OPC_SERVER' in os.environ:       opc_server = environ['OPC_SERVER']
    if 'OPC_GATE_HOST' in os.environ:    open_host = environ['OPC_GATE_HOST']
    if 'OPC_GATE_PORT' in os.environ:    open_port = int(environ['OPC_GATE_PORT'])
    if 'OPC_TIMEOUT' in os.environ:      timeout = int(environ['OPC_TIMEOUT'])

    # Parse command line arguments

    if argv.count('-') > 0:
        argv[argv.index('-')] = '--pipe'
        pipe = True

    try:
        opts, args = gnu_getopt(argv[1:], 'rwlpfiqRSevx:m:C:H:P:c:h:s:L:F:z:o:a:u:t:g:y:n:', ['read','write','list','properties','flat','info','mode=','gate-host=','gate-port=','class=','host=','server=','output=','pause=','pipe','servers','sessions','repeat=','function=','append=','update=','timeout=','size=','source=','id=','verbose','recursive','rotate=','errors','name='])
    except GetoptError:
        usage()
        exit()   

    for o, a in opts:
        if o in ['-m', '--mode']       : opc_mode = a
        if o in ['-C', '--class']      : opc_class = a
        if o in ['-n', '--name']       : client_name = a
        if o in ['-H', '--open-host']  : open_host = a;  opc_mode = 'open' 
        if o in ['-P', '--open-port']  : open_port = a;  opc_mode = 'open'
        if o in ['-h', '--host']       : opc_host = a
        if o in ['-s', '--server']     : opc_server = a

        if o in ['-r', '--read']       : action = 'read'
        if o in ['-w', '--write']      : action = 'write'
        if o in ['-l', '--list']       : action = 'list'
        if o in ['-f', '--flat']       : action = 'flat'
        if o in ['-p', '--properties'] : action = 'properties'
        if o in ['-i', '--info']       : action = 'info'
        if o in ['-q', '--servers']    : action = 'servers'
        if o in ['-S', '--sessions']   : action = 'sessions'

        if o in ['-o', '--output']     : style = a
        if o in ['-L', '--repeat']     : repeat_pause = float(a);
        if o in ['-F', '--function']   : read_function = a;
        if o in ['-z', '--pause']      : tx_pause = int(a)
        if o in ['-u', '--update']     : update_rate = int(a)
        if o in ['-t', '--timeout']    : timeout = int(a)
        if o in ['-g', '--size']       : group_size = int(a)
        if o in ['-c', '--source']     : data_source = a
        if o in ['-y', '--id']         : property_ids = a
        if o in ['-a', '--append']     : append = a
        if o in ['-x', '--rotate']     : num_columns = int(a)
        if o in ['-v', '--verbose']    : verbose = True
        if o in ['-e', '--errors']     : include_err_msg = True
        if o in ['-R', '--recursive']  : recursive = True
        if o in ['--pipe']             : pipe = True

    # Check validity of command line options

    if num_columns > 0 and style in ('values', 'pairs'):
        print("'%s' style format may not be used with rotate" % style)
        exit()
       
    if opc_mode not in ('open', 'dcom'):
        print("'%s' is not a valid protocol mode (options: dcom, open)" % opc_mode)
        exit()

    if opc_mode == 'dcom' and not OpenOPC.opcda.win32com_found:
        print("win32com modules required when using DCOM protocol mode (http://pywin32.sourceforge.net/)")
        exit()

    if opc_mode == 'open' and not pyro_found:
        print("Pyro module required when using Open protocol mode (http://pyro.sourceforge.net)")
        exit()

    if style not in ('table', 'values', 'pairs', 'csv', 'html'):
        print("'%s' is not a valid style format (options: table, values, pairs, csv, html)" % style)
        exit()

    if read_function not in ('sync', 'async'):
        print("'%s' is not a valid read function (options: sync, async)" % read_function)
        exit()
    else:
        sync = (read_function == 'sync')
       
    if data_source not in ('cache', 'device', 'hybrid'):
        print("'%s' is not a valid data source mode (options: cache, device, hybrid)" % data_source)
        exit()

    if len(argv[1:]) == 0 or argv[1] == '/?' or argv[1] == '--help':
        usage()
        exit()

    if opc_server == '' and action not in ('servers', 'sessions'):
       print('OPC server name missing: use -s option or set OPC_SERVER environment variable')
       exit()

    if data_source in ('cache', 'hybrid') and read_function == 'async' and update_rate == None and repeat_pause != None:
        update_rate = int(repeat_pause * 1000.0)
    elif update_rate == None:
        update_rate = -1

    # Build tag list

    tags = []

    # Tag list passed via standrd input
    if pipe:
        try:
            reader = csv.reader(sys.stdin)
            tags_nested = list(reader)
        except KeyboardInterrupt:
            exit()

        tags = [line[0] for line in tags_nested if len(line) > 0]
        if len(tags) == 0:
            print('Input stream must contain ITEMs (one per line)')
            exit()
          
        if action == 'write':
            try:
                tag_value_pairs = [(item[0], item[1]) for item in tags_nested]
            except IndexError:
                print('Write input must be in ITEM,VALUE (CSV) format')
                exit()

    # Tag list passed via command line arguments
    else:
        for a in args:
            tags.append(a.replace('+', ' '))
        tags_nested = [[tag] for tag in tags]

        if action == 'write':
            if len(tags) % 2 == 0:
                tag_value_pairs = [(tags[i], tags[i+1]) for i in range(0, len(tags), 2)]
            else:
                print('Write arguments must be supplied in ITEM=VALUE or ITEM VALUE format')
                exit()

    if len(append) > 0:
        tags = [t + a  for t in tags for a in append.split(',')]

    if property_ids != None:
        try:
            property_ids = [int(p) for p in property_ids.split(',')]
        except ValueError:
            print('Property ids must be numeric')
            exit()
       
    if action in ('read','write') and not pipe and len(tags) == 0:
        usage()
        exit()

    # Were only health monitoring "@" tags supplied?

    health_tags = [t for t in tags if t[:1] == '@']
    opc_tags = [t for t in tags if t[:1] != '@']
    if len(health_tags) > 0 and len(opc_tags) == 0:
        health_only = True
    else:
        health_only = False

    # Establish signal handler for keyboard interrupts

    sh = SigHandler()
    signal.signal(signal.SIGINT,sh)
    if os.name == 'nt':
        signal.signal(signal.SIGBREAK,sh)
    signal.signal(signal.SIGTERM,sh)

    # ACTION: List active sessions in OpenOPC service

    if action == 'sessions':
        print('  %-38s %-18s %-18s' % ('Remote Client', 'Start Time', 'Last Transaction'))
        try:
            for guid, host, init_time, tx_time in OpenOPC.get_sessions(open_host, open_port):
                print('  %-38s %-18s %-18s'  % (host, time2str(init_time), time2str(tx_time)))
        except:
            error_msg = sys.exc_info()[1]
            print("Cannot connect to OpenOPC service at %s:%s - %s" % (open_host, open_port, error_msg))
        exit()
       
    # Connect to OpenOPC service (Open mode)

    if opc_mode == 'open':
        try:
            opc = OpenOPC.open_client(open_host, open_port)
        except:
            error_msg = sys.exc_info()[1]
            print("Cannot connect to OpenOPC Gateway Service at %s:%s - %s" % (open_host, open_port, error_msg))
            exit()

    # Dispatch to COM class (DCOM mode)

    else:
        try:
            opc = OpenOPC.client(opc_class, client_name)
        except Exception as error_msg:
            print("Failed to initialize an OPC Automation Class from the search list '%s' - %s" % (opc_class, error_msg))
            exit()

    # Connect to OPC server

    if action not in ['servers'] and not health_only:
        try:
            opc.connect(opc_server, opc_host)
        except Exception as error_msg:
            if opc_mode == 'open': error_msg = error_msg[0]
            print("Connect to OPC server '%s' on '%s' failed - %s" % (opc_server, opc_host, error_msg))
            exit()

    # Perform requested action...

    start_time = time.time()

    # ACTION: Read Items

    if action == 'read':
        if group_size and len(tags) > group_size and opc_mode == 'dcom':
            opc_read = opc.iread
            rotate = irotate
        else:
            opc_read = opc.read
            rotate = rotatelist
               
        if verbose:
            def trace(msg): print(msg)
            opc.set_trace(trace)

        success_count = 0
        total_count = 0
        com_connected = True
        pyro_connected = True

        while not sh.signaled:

            try:
                if not pyro_connected:
                    opc = OpenOPC.open_client(open_host, open_port)
                    opc.connect(opc_server, opc_host)
                    opc_read = opc.read
                    pyro_connected = True
                    com_connected = True

                if not com_connected:
                    opc.connect(opc_server, opc_host)
                    com_connected = True

                status = output(rotate(opc_read(tags,
                                        group='test',
                                        size=group_size,
                                        pause=tx_pause,
                                        source=data_source,
                                        update=update_rate,
                                        timeout=timeout,
                                        sync=sync,
                                        include_error=include_err_msg),
                            num_columns), style)
                            
            except OpenOPC.TimeoutError as error_msg:
                if opc_mode == 'open': error_msg = error_msg[0]
                print(error_msg)
                success = False

            except Exception as error_msg:
                if opc_mode == 'open': error_msg = error_msg[0]
                print(error_msg)
                success = False

                if opc.ping():
                    com_connected = True
                else:
                    com_connected = False
                
            except (Pyro5.errors.ConnectionClosedError, Pyro5.errors.ProtocolError) as error_msg:
                print('Gateway Service: %s' % error_msg)
                success = False
                pyro_connected = False

            except TypeError as error_msg:
                if opc_mode == 'open': error_msg = error_msg #[0]
                print(error_msg)
                break

            except IOError:
                opc.close()
                exit()
                
            else:
                success = True

            if success and num_columns == 0:
                success_count += len([s for s in status if s[2] != 'Error'])
                total_count += len(status)

            if repeat_pause != None:
                try:
                    time.sleep(repeat_pause)
                except IOError:
                    break
            else:
                break

        if style == 'table' and num_columns == 0:
            print('\nRead %d of %d items (%.2f seconds)' % (success_count, total_count, time.time() - start_time))

        try:
            opc.remove('test')
        except Exception as error_msg:
            if opc_mode == 'open': error_msg = error_msg[0]
            print(error_msg)
             
    # ACTION: Write Items

    elif action == 'write':
        if group_size and len(tags) > group_size and opc_mode == 'dcom':
            opc_write = opc.iwrite
            rotate = irotate
        else:
            opc_write = opc.write
            rotate = rotatelist

        try:
            status = output(rotate(opc_write(tag_value_pairs,
                                      size=group_size,
                                      pause=tx_pause,
                                      include_error=include_err_msg),
                               num_columns), style)

        except OpenOPC.OPCError as error_msg:
            if opc_mode == 'open': error_msg = error_msg[0]
            print(error_msg)

        if style == 'table' and num_columns == 0:
            success = len([s for s in status if s[1] != 'Error'])
            print('\nWrote %d of %d items (%.2f seconds)' % (success, len(tag_value_pairs), time.time() - start_time))

    # ACTION: List Items (Tree Browser)
       
    elif action == 'list':
        if opc_mode == 'open':
            opc_list = opc.list
            rotate = rotatelist
        else:
            opc_list = opc.ilist
            rotate = irotate

        try:
            output(rotate(opc_list(tags, recursive=recursive), num_columns), style)
        except Exception as error_msg:
            if opc_mode == 'open': error_msg = error_msg[0]
            print(error_msg)

    # ACTION: List Items (Flat Browser)
       
    elif action == 'flat':
        try:
            output(opc.list(tags, flat=True), style)
        except Exception as error_msg:
            if opc_mode == 'open': error_msg = error_msg[0]
            print(error_msg)

    # ACTION: Item Properties

    elif action == 'properties':
        if opc_mode == 'open':
            opc_properties = opc.properties
            rotate = rotatelist
        else:
            opc_properties = opc.iproperties
            rotate = irotate

        if property_ids != None:
            value_idx = 2
        else:
            value_idx = 3

        try:
            output(rotate(opc_properties(tags, property_ids), num_columns, value_idx), style, value_idx)
        except Exception as error_msg:
            if opc_mode == 'open': error_msg = error_msg[0]
            print(error_msg)

    # ACTION: Server Info

    elif action == 'info':
        try:
            output(rotatelist(opc.info(), num_columns), style)
        except Exception as error_msg:
            if opc_mode == 'open': error_msg = error_msg[0]
            print(error_msg)

    # ACTION: List Servers

    elif action == 'servers':
        try:
            output(rotatelist(opc.servers(opc_host), num_columns), style)
        except Exception as error_msg:
            if opc_mode == 'open': error_msg = error_msg[0]
            print("Error getting server list from '%s' - %s" % (opc_host, error_msg))

    # Disconnect from OPC Server

    try:
        opc.close()
    except Exception as error_msg:
        if opc_mode == 'open': error_msg = error_msg[0]
        print(error_msg)

if __name__ == "__main__":
    main()


###########################################################################
#
# OpenOPC for Python OPC-DA Library File
#
# General functions file.
#
# Copyright (c) 2007-2012 Barry Barnreiter (barry_b@users.sourceforge.net)
# Copyright (c) 2014 Anton D. Kachalov (mouse@yandex.ru)
# Copyright (c) 2017 José A. Maita (jose.a.maita@gmail.com)
# Copyright (c) 2022 j3mg
#
###########################################################################
import os

try:
    import pythoncom # Only used by get_error_str on Python 32-bit systems
except:
    pass


# OPC Constants
ACCESS_RIGHTS = (0, 'Read', 'Write', 'Read/Write')
BROWSER_TYPE = (0, 'Hierarchical', 'Flat')
OPC_CLASS = 'OPC DA Automation Wrapper 2.02;Matrikon.OPC.Automation;Graybox.OPC.DAWrapper;HSCOPC.Automation;RSI.OPCAutomation;OPC.Automation'
OPC_SERVER = 'Hci.TPNServer;HwHsc.OPCServer;opc.deltav.1;AIM.OPC.1;Yokogawa.ExaopcDAEXQ.1;OSI.DA.1;OPC.PHDServerDA.1;Aspen.Infoplus21_DA.1;National Instruments.OPCLabVIEW;RSLinx OPC Server;KEPware.KEPServerEx.V4;Matrikon.OPC.Simulation;Prosys.OPC.Simulation;CCOPC.XMLWrapper.1;OPC.SimaticHMI.CoRtHmiRTm.1'
OPC_CLIENT = 'OpenOPC'
OPC_QUALITY = ('Bad', 'Uncertain', 'Unknown', 'Good')
OPC_STATUS = (0, 'Running', 'Failed', 'NoConfig', 'Suspended', 'Test')
SOURCE_CACHE = 1
SOURCE_DEVICE = 2

def exceptional(func, alt_return=None, alt_exceptions=(Exception,), final=None, catch=None):
   """Turns exceptions into an alternative return value"""

   def _exceptional(*args, **kwargs):
      try:
         try:
            return func(*args, **kwargs)
         except alt_exceptions:
            return alt_return
         except:
            if catch: return catch(sys.exc_info(), lambda:func(*args, **kwargs))
            raise
      finally:
         if final: final()
   return _exceptional

def get_error_str(error, _opc=None):
    """Return the error string for a OPC or COM error code"""

    hr, msg, exc, arg = error.args

    if exc == None:
        error_str = str(msg)
    else:
        scode = exc[5]

        if _opc is not None:
            try:
                opc_err_str = unicode(_opc.GetErrorString(scode)).strip('\r\n')
            except:
                opc_err_str = None
        else:
            opc_err_str = None

        try:
            com_err_str = unicode(pythoncom.GetScodeString(scode)).strip('\r\n')
        except:
            com_err_str = None

        # OPC error codes and COM error codes are overlapping concepts,
        # so we combine them together into a single error message.

        if opc_err_str == None and com_err_str == None:
            error_str = str(scode)
        elif opc_err_str == com_err_str:
            error_str = opc_err_str
        elif opc_err_str == None:
            error_str = com_err_str
        elif com_err_str == None:
            error_str = opc_err_str
        else:
            error_str = '%s (%s)' % (opc_err_str, com_err_str)

    return error_str

def quality_str(quality_bits):
    """Convert OPC quality bits to a descriptive string"""

    quality = (quality_bits >> 6) & 3
    return OPC_QUALITY[quality]

def type_check(tags):
    """Perform a type check on a list of tags"""

    if type(tags) in (list, tuple):
        single = False
    elif tags == None:
        tags = []
        single = False
    else:
        tags = [tags]
        single = True

    if len([t for t in tags if type(t) not in (str,bytes)]) == 0:
        valid = True
    else:
        valid = False

    return tags, single, valid

def tags2trace(tags):
    """Convert a list tags into a formatted string suitable for the trace callback log"""
    arg_str = ''
    for i,t in enumerate(tags[1:]):
        if i > 0: arg_str += ','
        arg_str += '%s' % t
    return arg_str

def wild2regex(string):
   """Convert a Unix wildcard glob into a regular expression"""
   return string.replace('.','[.]').replace('*','.*').replace('?','.').replace('!','^')

def TimeoutError(msg):
    return Exception("TimeoutError", msg)

def OPCError(msg):
    return Exception("OPCError", msg) # Pyro5 cannot serialize custom exception types



###########################################################################
#
# OpenOPC Service for Python Library Module
#
# Copyright (c) 2007-2012 Barry Barnreiter (barry_b@users.sourceforge.net)
# Copyright (c) 2014 Anton D. Kachalov (mouse@yandex.ru)
# Copyright (c) 2017 José A. Maita (jose.a.maita@gmail.com)
# Copyright (c) 2022 j3mg
#
###########################################################################

# Main service class
from .opcservice import opc as opc



###########################################################################
#
# OpenOPC for Python OPC-DA Library File
#
# A Windows only OPC-DA library file.
#
# Copyright (c) 2007-2012 Barry Barnreiter (barry_b@users.sourceforge.net)
# Copyright (c) 2014 Anton D. Kachalov (mouse@yandex.ru)
# Copyright (c) 2017 José A. Maita (jose.a.maita@gmail.com)
# Copyright (c) 2022 j3mg
#
###########################################################################
import os
import sys
import time
import types
import string
import socket
import re
import Pyro5.core
import Pyro5.server

# OPC Constants
from OpenOPC.common import get_error_str, OPC_CLASS, OPC_SERVER, OPC_CLIENT, OPCError

win32com_found = False
win32_found = (os.name == 'nt')

# Win32 only modules not needed for 'open' protocol mode
if win32_found:
    try:
        import win32com.client
        import win32com.server.util
        import win32event
        import pythoncom
        import pywintypes
        import OpenOPC.systemhealth

        # Win32 variant types
        pywintypes.datetime = pywintypes.TimeType
        vt = dict([(pythoncom.__dict__[vtype], vtype) for vtype in pythoncom.__dict__.keys() if vtype[:2] == "VT"])

        # Allow gencache to create the cached wrapper objects
        win32com.client.gencache.is_readonly = False

        # Under p2exe the call in gencache to __init__() does not happen
        # so we use Rebuild() to force the creation of the gen_py folder
        win32com.client.gencache.Rebuild(verbose=0)

    # So we can work on Windows in "open" protocol mode without the need for the win32com modules
    except ImportError:
        win32com_found = False
    else:
        from OpenOPC.opcdaio import ClientIO
        from OpenOPC.opcdatools import ClientTools
        
        win32com_found = True

@Pyro5.server.expose    # needed for 5.12
class client():
    def __init__(self, opc_class=None, client_name=None):
        """Instantiate OPC automation class"""
        
        self.opc_server = None
        self.opc_host = None
        self.client_name = client_name
        self._groups = {}
        self._group_tags = {}
        self._group_valid_tags = {}
        self._group_server_handles = {}
        self._group_handles_tag = {}
        self._group_hooks = {}
        # Members prefixed by __open are owned by the OPC Gateway Service
        self.__open_serv__ = None 
        self.__open_self__ = None
        self.__open_host__ = None
        self.__open_port__ = None
        self.__open_guid__ = None

        self.trace = None

        def win32_init(opc_class):
            pythoncom.CoInitialize()

            if opc_class == None:
                if 'OPC_CLASS' in os.environ:
                    opc_class = os.environ['OPC_CLASS']
                else:
                    opc_class = OPC_CLASS

            opc_class_list = opc_class.split(';')

            for i,c in enumerate(opc_class_list):
                try:
                    self._opc = win32com.client.gencache.EnsureDispatch(c, 0)
                    self.opc_class = c
                    break
                except pythoncom.com_error as err:
                    if i == len(opc_class_list)-1:
                        error_msg = 'Dispatch: %s' % get_error_str(err)
                        raise OPCError(error_msg)
            self._event = win32event.CreateEvent(None,0,0,None)

            self.clientIO = ClientIO()
            self.clientTools = ClientTools(self.opc_class, self.opc_host)

        self.win32os = win32com_found # win32com_found set to False by pytest-mock in test_client
        if self.win32os: win32_init(opc_class)

    def set_gateway_settings(self, service, instance, host, port, guid):
        """ For the use of the OPC Gateway Service """
        
        self.__open_serv__ = service 
        self.__open_self__ = instance
        self.__open_host__ = host
        self.__open_port__ = port
        self.__open_guid__ = guid
        if self.win32os: self.clientTools.set_gateway_settings(service, host, port, guid)

    def set_trace(self, trace):
        if self.__open_serv__ == None:
            self.trace = trace
            if self.win32os: self.clientIO.setTrace(trace)

    def connect(self, opc_server=None, opc_host='localhost'):
        """Connect to the specified OPC server"""
        
        def win32_connect(opc_server, opc_host):
            pythoncom.CoInitialize()
            if opc_server == None:
                # Initial connect using environment vars
                if self.opc_server == None:
                    if 'OPC_SERVER' in os.environ:
                        opc_server = os.environ['OPC_SERVER']
                    else:
                        opc_server = OPC_SERVER
                # Reconnect using previous server name
                else:
                    opc_server = self.opc_server
                    opc_host = self.opc_host

            opc_server_list = opc_server.split(';')
            connected = False

            for s in opc_server_list:
                if len(s):
                    try:
                        if self.trace: self.trace('Connect(%s,%s)' % (s, opc_host))
                        self._opc.Connect(s, opc_host)
                    except pythoncom.com_error as err:
                        if len(opc_server_list) == 1:
                            error_msg = 'Connect: %s' % get_error_str(err)
                            raise OPCError(error_msg)
                    else:
                        # Set client name since some OPC servers use it for security
                        try:
                            if self.client_name == None:
                                if 'OPC_CLIENT' in os.environ:
                                    self._opc.ClientName = os.environ['OPC_CLIENT']
                                else:
                                    self._opc.ClientName = OPC_CLIENT
                            else:
                                self._opc.ClientName = self.client_name
                        except:
                            pass
                        connected = True
                        break

            if not connected:
                raise OPCError('Connect: Cannot connect to any of the servers in the OPC_SERVER list')

            # With some OPC servers, the next OPC call immediately after Connect()
            # will occationally fail.  Sleeping for 1/100 second seems to fix this.
            time.sleep(0.01)

            self.opc_server = opc_server
            if opc_host == 'localhost':
                opc_host = socket.gethostname()
            self.opc_host = opc_host
            self.clientTools.set_opc_host(opc_host)
            return connected
        
        connected = False
        if self.win32os:
            connected = win32_connect(opc_server, opc_host)
        return connected

    def GUID(self):
        return self.__open_guid__

    def close(self, del_object=True):
        """Disconnect from the currently connected OPC server"""

        def win32_close(del_object):
            try:
                pythoncom.CoInitialize()
                self.remove(self.groups())

            except pythoncom.com_error as err:
                error_msg = 'Disconnect: %s' % get_error_str(err, self._opc)
                raise OPCError(error_msg)

            except OPCError:
                pass

            finally:
                if self.trace: self.trace('Disconnect()')
                self._opc.Disconnect()

                # Remove this object from the open gateway service
                if self.__open_serv__ and del_object:
                    self.__open_serv__.release_client(self.__open_self__)

        if self.win32os: win32_close(del_object)

    def groups(self):
        """Return a list of active tag groups"""
        return self._groups.keys()


    #
    # Read/Write functions
    #
    def iread(self, tags=None, group=None, size=None, pause=0, source='hybrid', update=-1, timeout=5000, sync=False, include_error=False, rebuild=False):
        if self.win32os:
            return self.clientIO.iread(self._opc, self.clientTools, tags, group, size, pause, source, update, timeout, sync, include_error, rebuild)
        else:
            return None

    def read(self, tags=None, group=None, size=None, pause=0, source='hybrid', update=-1, timeout=5000, sync=False, include_error=False, rebuild=False):
        if self.win32os:
            return self.clientIO.read(self._opc, self.clientTools, tags, group, size, pause, source, update, timeout, sync, include_error, rebuild)
        else:
            return None

    def _read_health(self, tags):
        if self.win32os:
            return self.clientIO._read_health(self.clientTools, tags)
        else:
            return None

    def iwrite(self, tag_value_pairs, size=None, pause=0, include_error=False):
        if self.win32os:
            return self.clientIO.iwrite(self._opc, self.clientTools, tag_value_pairs, size, pause, include_error)
        else:
            return None

    def write(self, tag_value_pairs, size=None, pause=0, include_error=False):
        if self.win32os:
            return self.clientIO.write(self._opc, self.clientTools, tag_value_pairs, size, pause, include_error)
        else:
            return None

    def remove(self, groups):
        if self.win32os:
            return self.clientIO.remove(self._opc, groups)
        else:
            return None

    def __getitem__(self, key):
        if self.win32os:
            return self.clientIO.__getitem__(self._opc, self.clientTools, key)
        else:
            return None

    def __setitem__(self, key, value):
        if self.win32os:
            return self.clientIO.__setitem__(self._opc, self.clientTools, key, value)
        else:
            return None
    #
    # General tools functions
    #
    def iproperties(self, tags, id=None):
        if self.win32os:
            return self.clientTools.iproperties(self._opc, tags, id)
        else:
            return None

    def properties(self, tags, id=None):
        if self.win32os:
            return self.clientTools.properties(self._opc, tags, id)
        else:
            return None

    def ilist(self, paths='*', recursive=False, flat=False, include_type=False):
        if self.win32os:
            return self.clientTools.ilist(self._opc, paths, recursive, flat, include_type)
        else:
            return None

    def list(self, paths='*', recursive=False, flat=False, include_type=False):
        if self.win32os:
            return self.clientTools.list(self._opc, paths, recursive, flat, include_type)
        else:
            return None

    def servers(self, opc_host='localhost'):
        if self.win32os:
            return self.clientTools.servers(self._opc, opc_host)
        else:
            return None

    def info(self):
        if self.win32os:
            return self.clientTools.info(self._opc)
        else:
            return None

    def ping(self):
        if self.win32os:
            return self.clientTools.ping(self._opc)
        else:
            return None

    def _get_error_str(self, err):
        if self.win32os:
            return get_error_str(err, self._opc)
        else:
            return None

    def _update_tx_time(self):
        if self.win32os:
            return self.clientTools._update_tx_time()
        else:
            return False



###########################################################################
#
# OpenOPC for Python OPC-DA IO Library file
#
# A Windows only OPC-DA library file.
#
# Copyright (c) 2007-2012 Barry Barnreiter (barry_b@users.sourceforge.net)
# Copyright (c) 2014 Anton D. Kachalov (mouse@yandex.ru)
# Copyright (c) 2017 José A. Maita (jose.a.maita@gmail.com)
# Copyright (c) 2022 j3mg
#
###########################################################################
import re
import OpenOPC.systemhealth
import win32com.client
import win32com.server.util
import win32event
import pythoncom
import pywintypes
import time
from multiprocessing import Queue
from OpenOPC.common import get_error_str, quality_str, type_check, tags2trace, TimeoutError, OPCError, SOURCE_CACHE, SOURCE_DEVICE, OPC_QUALITY

current_client = None

class GroupEvents:
    def __init__(self):
        self.client = current_client

    def OnDataChange(self, TransactionID, NumItems, ClientHandles, ItemValues, Qualities, TimeStamps):
        self.client.callback_queue.put((TransactionID, ClientHandles, ItemValues, Qualities, TimeStamps))

class ClientIO():
    def __init__(self):
        self.trace = None

        self.callback_queue = Queue()

        self._tx_id = 0
        # On reconnect we need to remove the old group names from OpenOPC's internal
        # cache since they are now invalid
        self._groups = {}
        self._group_tags = {}
        self._group_valid_tags = {}
        self._group_server_handles = {}
        self._group_handles_tag = {}
        self._group_hooks = {}
        self.cpu = None

    def setTrace(self, trace):
        self.trace = trace

    def iread(self, _opc, clientTools, tags=None, group=None, size=None, pause=0, source='hybrid', update=-1, timeout=5000, sync=False, include_error=False, rebuild=False):
        """Iterable version of read()"""

        def add_items(tags):
            names = list(tags)

            names.insert(0,0)
            errors = []

            if self.trace: self.trace('Validate(%s)' % tags2trace(names))

            try:
                errors = opc_items.Validate(len(names)-1, names)
            except:
                pass

            valid_tags = []
            valid_values = []
            client_handles = []

            if not sub_group in self._group_handles_tag:
                self._group_handles_tag[sub_group] = {}
                n = 0
            elif len(self._group_handles_tag[sub_group]) > 0:
                n = max(self._group_handles_tag[sub_group]) + 1
            else:
                n = 0

            for i, tag in enumerate(tags):
                if errors[i] == 0:
                    valid_tags.append(tag)
                    client_handles.append(n)
                    self._group_handles_tag[sub_group][n] = tag
                    n += 1
                elif include_error:
                    error_msgs[tag] = _opc.GetErrorString(errors[i])

                if self.trace and errors[i] != 0: self.trace('%s failed validation' % tag)

            client_handles.insert(0,0)
            valid_tags.insert(0,0)
            server_handles = []
            errors = []

            if self.trace: self.trace('AddItems(%s)' % tags2trace(valid_tags))

            try:
                server_handles, errors = opc_items.AddItems(len(client_handles)-1, valid_tags, client_handles)
            except:
                pass

            valid_tags_tmp = []
            server_handles_tmp = []
            valid_tags.pop(0)

            if not sub_group in self._group_server_handles:
                self._group_server_handles[sub_group] = {}

            for i, tag in enumerate(valid_tags):
                if errors[i] == 0:
                    valid_tags_tmp.append(tag)
                    server_handles_tmp.append(server_handles[i])
                    self._group_server_handles[sub_group][tag] = server_handles[i]
                elif include_error:
                    error_msgs[tag] = _opc.GetErrorString(errors[i])

            valid_tags = valid_tags_tmp
            server_handles = server_handles_tmp

            return valid_tags, server_handles

        def remove_items(tags):
            if self.trace: self.trace('RemoveItems(%s)' % tags2trace(['']+tags))
            server_handles = [self._group_server_handles[sub_group][tag] for tag in tags]
            server_handles.insert(0,0)
            errors = []

            try:
                errors = opc_items.Remove(len(server_handles)-1, server_handles)
            except pythoncom.com_error as err:
                error_msg = 'RemoveItems: %s' % get_error_str(err, _opc)
                raise OPCError(error_msg)

        try:
            clientTools._update_tx_time()
            pythoncom.CoInitialize()

            if include_error:
                sync = True

            if sync:
                update = -1

            tags, single, valid = type_check(tags)
            if not valid:
                raise TypeError("iread(): 'tags' parameter must be a string or a list of strings")

            # Group exists
            if group in self._groups and not rebuild:
                num_groups = self._groups[group]
                data_source = SOURCE_CACHE

            # Group non-existant
            else:
                if size:
                    # Break-up tags into groups of 'size' tags
                    tag_groups = [tags[i:i+size] for i in range(0, len(tags), size)]
                else:
                    tag_groups = [tags]

                num_groups = len(tag_groups)
                data_source = SOURCE_DEVICE

            results = []

            for gid in range(num_groups):
                if gid > 0 and pause > 0: time.sleep(pause/1000.0)

                error_msgs = {}
                opc_groups = _opc.OPCGroups
                opc_groups.DefaultGroupUpdateRate = update

                # Anonymous group
                if group == None:
                    try:
                        if self.trace: self.trace('AddGroup()')
                        opc_group = opc_groups.Add()
                    except pythoncom.com_error as err:
                        error_msg = 'AddGroup: %s' % get_error_str(err, _opc)
                        raise OPCError(error_msg)
                    sub_group = group
                    new_group = True
                else:
                    sub_group = '%s.%d' % (group, gid)

                    # Existing named group
                    try:
                        if self.trace: self.trace('GetOPCGroup(%s)' % sub_group)
                        opc_group = opc_groups.GetOPCGroup(sub_group)
                        new_group = False

                    # New named group
                    except:
                        try:
                            if self.trace: self.trace('AddGroup(%s)' % sub_group)
                            opc_group = opc_groups.Add(sub_group)
                        except pythoncom.com_error as err:
                            error_msg = 'AddGroup: %s' % get_error_str(err, _opc)
                            raise OPCError(error_msg)
                        self._groups[str(group)] = len(tag_groups)
                        new_group = True

                opc_items = opc_group.OPCItems

                if new_group:
                    opc_group.IsSubscribed = 1
                    opc_group.IsActive = 1
                    if not sync:
                        if self.trace: self.trace('WithEvents(%s)' % opc_group.Name)
                        global current_client
                        current_client = self
                        self._group_hooks[opc_group.Name] = win32com.client.WithEvents(opc_group, GroupEvents)

                    tags = tag_groups[gid]

                    valid_tags, server_handles = add_items(tags)

                    self._group_tags[sub_group] = tags
                    self._group_valid_tags[sub_group] = valid_tags

                # Rebuild existing group
                elif rebuild:
                    tags = tag_groups[gid]

                    valid_tags = self._group_valid_tags[sub_group]
                    add_tags = [t for t in tags if t not in valid_tags]
                    del_tags = [t for t in valid_tags if t not in tags]

                    if len(add_tags) > 0:
                        valid_tags, server_handles = add_items(add_tags)
                        valid_tags = self._group_valid_tags[sub_group] + valid_tags

                    if len(del_tags) > 0:
                        remove_items(del_tags)
                        valid_tags = [t for t in valid_tags if t not in del_tags]

                    self._group_tags[sub_group] = tags
                    self._group_valid_tags[sub_group] = valid_tags

                    if source == 'hybrid': data_source = SOURCE_DEVICE

                # Existing group
                else:
                    tags = self._group_tags[sub_group]
                    valid_tags = self._group_valid_tags[sub_group]
                    if sync:
                        server_handles = [item.ServerHandle for item in opc_items]

                tag_value = {}
                tag_quality = {}
                tag_time = {}
                tag_error = {}

                # Sync Read
                if sync:
                    values = []
                    errors = []
                    qualities = []
                    timestamps= []

                    if len(valid_tags) > 0:
                        server_handles.insert(0,0)

                        if source != 'hybrid':
                            data_source = SOURCE_CACHE if source == 'cache' else SOURCE_DEVICE

                        if self.trace: self.trace('SyncRead(%s)' % data_source)

                        try:
                            values, errors, qualities, timestamps = opc_group.SyncRead(data_source, len(server_handles)-1, server_handles)
                        except pythoncom.com_error as err:
                            error_msg = 'SyncRead: %s' % get_error_str(err, _opc)
                            raise OPCError(error_msg)

                        for i,tag in enumerate(valid_tags):
                            tag_value[tag] = values[i]
                            tag_quality[tag] = qualities[i]
                            tag_time[tag] = timestamps[i]
                            tag_error[tag] = errors[i]

                # Async Read
                else:
                    if len(valid_tags) > 0:
                        if self._tx_id >= 0xFFFF:
                            self._tx_id = 0
                        self._tx_id += 1

                        if source != 'hybrid':
                            data_source = SOURCE_CACHE if source == 'cache' else SOURCE_DEVICE

                        if self.trace: self.trace('AsyncRefresh(%s)' % data_source)

                        try:
                            opc_group.AsyncRefresh(data_source, self._tx_id)
                        except pythoncom.com_error as err:
                            error_msg = 'AsyncRefresh: %s' % get_error_str(err, _opc)
                            raise OPCError(error_msg)

                        tx_id = 0
                        start = time.time() * 1000

                        while tx_id != self._tx_id:
                            now = time.time() * 1000
                            if now - start > timeout:
                                raise TimeoutError('Callback: Timeout waiting for data')

                            if self.callback_queue.empty():
                                pythoncom.PumpWaitingMessages()
                            else:
                                tx_id, handles, values, qualities, timestamps = self.callback_queue.get()

                        for i,h in enumerate(handles):
                            tag = self._group_handles_tag[sub_group][h]
                            tag_value[tag] = values[i]
                            tag_quality[tag] = qualities[i]
                            tag_time[tag] = timestamps[i]

                for tag in tags:
                    if tag in tag_value:
                        if (not sync and len(valid_tags) > 0) or (sync and tag_error[tag] == 0):
                            value = tag_value[tag]
                            if type(value) == pywintypes.TimeType:
                                value = str(value)
                            quality = quality_str(tag_quality[tag])
                            timestamp = str(tag_time[tag])
                        else:
                            value = None
                            quality = 'Error'
                            timestamp = None
                        if include_error:
                            error_msgs[tag] = _opc.GetErrorString(tag_error[tag]).strip('\r\n')
                    else:
                        value = None
                        quality = 'Error'
                        timestamp = None
                        if include_error and not tag in error_msgs:
                            error_msgs[tag] = ''

                    if single:
                        if include_error:
                            yield (value, quality, timestamp, error_msgs[tag])
                        else:
                            yield (value, quality, timestamp)
                    else:
                        if include_error:
                            yield (tag, value, quality, timestamp, error_msgs[tag])
                        else:
                            yield (tag, value, quality, timestamp)

                if group == None:
                    try:
                        if not sync and opc_group.Name in self._group_hooks:
                            if self.trace: self.trace('CloseEvents(%s)' % opc_group.Name)
                            self._group_hooks[opc_group.Name].close()

                        if self.trace: self.trace('RemoveGroup(%s)' % opc_group.Name)
                        opc_groups.Remove(opc_group.Name)

                    except pythoncom.com_error as err:
                        error_msg = 'RemoveGroup: %s' % get_error_str(err, _opc)
                        raise OPCError(error_msg)

        except pythoncom.com_error as err:
            error_msg = 'read: %s' % get_error_str(err, _opc)
            raise OPCError(error_msg)

    def read(self, _opc, clientTools, tags=None, group=None, size=None, pause=0, source='hybrid', update=-1, timeout=5000, sync=False, include_error=False, rebuild=False):
        """Return list of (value, quality, time) tuples for the specified tag(s)"""

        tags_list, single, valid = type_check(tags)
        if not valid:
            raise TypeError("read(): 'tags' parameter must be a string or a list of strings")

        num_health_tags = len([t for t in tags_list if t[:1] == '@'])
        num_opc_tags = len([t for t in tags_list if t[:1] != '@'])

        if num_health_tags > 0:
            if num_opc_tags > 0:
                raise TypeError("read(): system health and OPC tags cannot be included in the same group")
            results = self._read_health(clientTools, tags)
        else:
            results = self.iread(_opc, clientTools, tags, group, size, pause, source, update, timeout, sync, include_error, rebuild)

        if single:
            return list(results)[0]
        else:
            return list(results)

    def _read_health(self, clientTools, tags):
        """Return values of special system health monitoring tags"""

        clientTools._update_tx_time()
        tags, single, valid = type_check(tags)

        time_str = time.strftime('%x %H:%M:%S')
        results = []

        for t in tags:
            if   t == '@MemFree':      value = OpenOPC.systemhealth.mem_free()
            elif t == '@MemUsed':      value = OpenOPC.systemhealth.mem_used()
            elif t == '@MemTotal':     value = OpenOPC.systemhealth.mem_total()
            elif t == '@MemPercent':   value = OpenOPC.systemhealth.mem_percent()
            elif t == '@DiskFree':     value = OpenOPC.systemhealth.disk_free()
            elif t == '@SineWave':     value = OpenOPC.systemhealth.sine_wave()
            elif t == '@SawWave':      value = OpenOPC.systemhealth.saw_wave()

            elif t == '@CpuUsage':
                if self.cpu == None:
                    self.cpu = OpenOPC.systemhealth.CPU()
                    time.sleep(0.1)
                value = self.cpu.get_usage()

            else:
                value = None

                m = re.match('@TaskMem(((.*?)))', t)
                if m:
                    image_name = m.group(1)
                    value = OpenOPC.systemhealth.task_mem(image_name)

                m = re.match('@TaskCpu(((.*?)))', t)
                if m:
                    image_name = m.group(1)
                    value = OpenOPC.systemhealth.task_cpu(image_name)

                m = re.match('@TaskExists(((.*?)))', t)
                if m:
                    image_name = m.group(1)
                    value = OpenOPC.systemhealth.task_exists(image_name)

            if value == None:
                quality = 'Error'
            else:
                quality = 'Good'

            if single:
                results.append((value, quality, time_str))
            else:
                results.append((t, value, quality, time_str))

        return results

    def iwrite(self, _opc, clientTools, tag_value_pairs, size=None, pause=0, include_error=False):
        """Iterable version of write()"""

        try:
            clientTools._update_tx_time()
            pythoncom.CoInitialize()

            def _valid_pair(p):
                if type(p) in (list, tuple) and len(p) >= 2 and type(p[0]) in (str,bytes):
                    return True
                else:
                    return False

            if type(tag_value_pairs) not in (list, tuple):
                raise TypeError("write(): 'tag_value_pairs' parameter must be a (tag, value) tuple or a list of (tag,value) tuples")

            if tag_value_pairs == None:
                tag_value_pairs = ['']
                single = False
            elif type(tag_value_pairs[0]) in (str,bytes):
                tag_value_pairs = [tag_value_pairs]
                single = True
            else:
                single = False

            invalid_pairs = [p for p in tag_value_pairs if not _valid_pair(p)]
            if len(invalid_pairs) > 0:
                raise TypeError("write(): 'tag_value_pairs' parameter must be a (tag, value) tuple or a list of (tag,value) tuples")

            names = [tag[0] for tag in tag_value_pairs]
            tags = [tag[0] for tag in tag_value_pairs]
            values = [tag[1] for tag in tag_value_pairs]

            # Break-up tags & values into groups of 'size' tags
            if size:
                name_groups = [names[i:i+size] for i in range(0, len(names), size)]
                tag_groups = [tags[i:i+size] for i in range(0, len(tags), size)]
                value_groups = [values[i:i+size] for i in range(0, len(values), size)]
            else:
                name_groups = [names]
                tag_groups = [tags]
                value_groups = [values]

            num_groups = len(tag_groups)

            status = []

            for gid in range(num_groups):
                if gid > 0 and pause > 0: time.sleep(pause/1000.0)

                opc_groups = _opc.OPCGroups
                opc_group = opc_groups.Add()
                opc_items = opc_group.OPCItems

                names = name_groups[gid]
                tags = tag_groups[gid]
                values = value_groups[gid]

                names.insert(0,0)
                errors = []

                try:
                    errors = opc_items.Validate(len(names)-1, names)
                except:
                    pass

                n = 1
                valid_tags = []
                valid_values = []
                client_handles = []
                error_msgs = {}

                for i, tag in enumerate(tags):
                    if errors[i] == 0:
                        valid_tags.append(tag)
                        valid_values.append(values[i])
                        client_handles.append(n)
                        error_msgs[tag] = ''
                        n += 1
                    elif include_error:
                        error_msgs[tag] = _opc.GetErrorString(errors[i])

                client_handles.insert(0,0)
                valid_tags.insert(0,0)
                server_handles = []
                errors = []

                try:
                    server_handles, errors = opc_items.AddItems(len(client_handles)-1, valid_tags, client_handles)
                except:
                    pass

                valid_tags_tmp = []
                valid_values_tmp = []
                server_handles_tmp = []
                valid_tags.pop(0)

                for i, tag in enumerate(valid_tags):
                    if errors[i] == 0:
                        valid_tags_tmp.append(tag)
                        valid_values_tmp.append(valid_values[i])
                        server_handles_tmp.append(server_handles[i])
                        error_msgs[tag] = ''
                    elif include_error:
                        error_msgs[tag] = _opc.GetErrorString(errors[i])

                valid_tags = valid_tags_tmp
                valid_values = valid_values_tmp
                server_handles = server_handles_tmp

                server_handles.insert(0,0)
                valid_values.insert(0,0)
                errors = []

                if len(valid_values) > 1:
                    try:
                        errors = opc_group.SyncWrite(len(server_handles)-1, server_handles, valid_values)
                    except:
                        pass

                n = 0
                for tag in tags:
                    if tag in valid_tags:
                        if errors[n] == 0:
                            status = 'Success'
                        else:
                            status = 'Error'
                        if include_error:  error_msgs[tag] = _opc.GetErrorString(errors[n])
                        n += 1
                    else:
                        status = 'Error'

                    # OPC servers often include newline and carriage return characters
                    # in their error message strings, so remove any found.
                    if include_error:  error_msgs[tag] = error_msgs[tag].strip('\r\n')

                    if single:
                        if include_error:
                            yield (status, error_msgs[tag])
                        else:
                            yield status
                    else:
                        if include_error:
                            yield (tag, status, error_msgs[tag])
                        else:
                            yield (tag, status)

                opc_groups.Remove(opc_group.Name)

        except pythoncom.com_error as err:
            error_msg = 'write: %s' % get_error_str(err, _opc)
            raise OPCError(error_msg)

    def write(self, _opc, clientTools, tag_value_pairs, size=None, pause=0, include_error=False):
        """Write list of (tag, value) pair(s) to the server"""

        if type(tag_value_pairs) in (list, tuple) and type(tag_value_pairs[0]) in (list, tuple):
            single = False
        else:
            single = True

        status = self.iwrite(_opc, clientTools, tag_value_pairs, size, pause, include_error)

        if single:
            return list(status)[0]
        else:
            return list(status)

    def remove(self, _opc, groups):
        """Remove the specified tag group(s)"""

        try:
            pythoncom.CoInitialize()
            opc_groups = _opc.OPCGroups

            if type(groups) in (str,bytes):
                groups = [groups]
                single = True
            else:
                single = False

            status = []
            groups_deleted = False

            for group in groups:
                if group in self._groups:
                    for i in range(self._groups[group]):
                        sub_group = '%s.%d' % (group, i)

                        if sub_group in self._group_hooks:
                            if self.trace: self.trace('CloseEvents(%s)' % sub_group)
                            self._group_hooks[sub_group].close()

                        try:
                            if self.trace: self.trace('RemoveGroup(%s)' % sub_group)
                            errors = opc_groups.Remove(sub_group)
                            if errors == None:
                                groups_deleted = True
                        except pythoncom.com_error as err:
                            error_msg = 'RemoveGroup: %s' % get_error_str(err, _opc)
                            raise OPCError(error_msg)

                        del(self._group_tags[sub_group])
                        del(self._group_valid_tags[sub_group])
                        del(self._group_handles_tag[sub_group])
                        del(self._group_server_handles[sub_group])
                    del(self._groups[group])
            return groups_deleted

        except pythoncom.com_error as err:
            error_msg = 'remove: %s' % get_error_str(err, _opc)
            raise OPCError(error_msg)

    def __getitem__(self, _opc, clientTools, key):
        """Read single item (tag as dictionary key)"""
        value, quality, time_str = self.read(_opc, clientTools, key)
        return value

    def __setitem__(self, _opc, clientTools, key, value):
        """Write single item (tag as dictionary key)"""
        self.write(_opc, clientTools, (key, value))
        return (key, value)



###########################################################################
#
# OpenOPC for Python Library Module
#
# Copyright (c) 2007-2012 Barry Barnreiter (barry_b@users.sourceforge.net)
# Copyright (c) 2014 Anton D. Kachalov (mouse@yandex.ru)
# Copyright (c) 2017 José A. Maita (jose.a.maita@gmail.com)
# Copyright (c) 2022 j3mg
#
###########################################################################

import os
from Pyro5 import __version__ as _pyro_version_
import Pyro5.core
import Pyro5.server
from OpenOPC.common import OPC_CLASS, OPC_CLIENT, OPC_SERVER, OPCError, TimeoutError # Required for the OpenOPC Gateway service
from OpenOPC.opcda import client as client

__version__ = '1.5.0'

def win32_check(): # can be mocked by pytest-mock
    return (os.name == 'nt')

def parse_version(ver):
    return tuple(map(int, (ver.split(".")))) # Works with Pyro5 numeric version strings


if parse_version("5.13.1") > parse_version(_pyro_version_):
    raise ImportError("Pyro version must be greater than 5.13")

def get_sessions(host='localhost', port=7766):
    """Return sessions in OpenOPC Gateway Service as GUID:host hash"""

    import Pyro5.client
    server_obj = Pyro5.client.Proxy("PYRO:opc@{0}:{1}".format(host, port))
    return server_obj.get_clients()

def open_client(host='localhost', port=7766):
    """Connect to the specified OpenOPC Gateway Service"""

    import Pyro5.client
    server_obj = Pyro5.client.Proxy("PYRO:opc@{0}:{1}".format(host, port))
    return server_obj.create_client()




###########################################################################
#
# OpenOPC Gateway Service
#
# A Windows service providing remote access to the OpenOPC library.
#
# Copyright (c) 2007-2012 Barry Barnreiter (barry_b@users.sourceforge.net)
# Copyright (c) 2014 Anton D. Kachalov (mouse@yandex.ru)
# Copyright (c) 2017 José A. Maita (jose.a.maita@gmail.com)
#
###########################################################################

import win32serviceutil
import win32service
import win32event
import servicemanager
import winerror
import select
import socket
import os
import sys
import time
import OpenOPC

try:
    import Pyro5.core
    import Pyro5.client
    import Pyro5.server
except ImportError:
    print('Pyro5 module required (https://pypi.python.org/pypi/Pyro5)')
    exit()

Pyro5.config.SERVERTYPE='thread'

opc_class = os.getenv('OPC_CLASS', OpenOPC.OPC_CLASS)
opc_gate_host = os.getenv('OPC_GATE_HOST', 'localhost')
opc_gate_port = int(os.getenv('OPC_GATE_PORT', 7766))


@Pyro5.server.expose    # needed for version 5.12
class opc(object):
    def __init__(self):
        self._remote_hosts = {}
        self._init_times = {}
        self._tx_times = {}

    def get_clients(self):
        """Return list of server instances as a list of (GUID,host,time) tuples"""
        hlist = []

        if self._pyroDaemon != None:
            reg1 = Pyro5.server.DaemonObject(self._pyroDaemon).registered()   # needed for version 5.12
            reg2 = [si for si in reg1 if si.find('obj_') == 0]
            reg = ["PYRO:{0}@{1}:{2}".format(obj, opc_gate_host, opc_gate_port) for obj in reg2]
            hosts = self._remote_hosts
            init_times = self._init_times
            tx_times = self._tx_times
            hlist = [(hosts[k] if k in hosts else '', init_times[k], tx_times[k]) for k in reg]

        return hlist

    def create_client(self, weakRegister=False): # Requires Pyro5 5.13 or above
        """Create a new OpenOPC instance in the Pyro server"""

        opc_obj = OpenOPC.client(opc_class)
        uri = self._pyroDaemon.register(opc_obj, weak=weakRegister)

        uuid = uri.__str__() # undocumented Pyro function
        opc_obj.set_gateway_settings(self, opc_obj, opc_gate_host, opc_gate_port, uuid)

        remote_ip = uuid # self.getLocalStorage().caller.addr[0]
        self._remote_hosts[uuid] = '%s' % (remote_ip)
        self._init_times[uuid] =  time.time()
        self._tx_times[uuid] =  time.time()
        return Pyro5.client.Proxy(uri)

    def release_client(self, obj):
        """Release an OpenOPC instance in the Pyro server"""
        """Called in client.close() method"""
        """REQUIRED unless object is created with weak register"""

        self._pyroDaemon.unregister(obj)
        del self._remote_hosts[obj.GUID()]
        del self._init_times[obj.GUID()]
        del self._tx_times[obj.GUID()]
        del obj

class OpcService(win32serviceutil.ServiceFramework):
    _svc_name_ = "zzzOpenOPCService"
    _svc_display_name_ = "OpenOPC Gateway Service"

    def __init__(self, args):
        win32serviceutil.ServiceFramework.__init__(self, args)
        self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)

    def SvcStop(self):
        servicemanager.LogInfoMsg('\n\nStopping service')
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        win32event.SetEvent(self.hWaitStop)

    def SvcDoRun(self):
        servicemanager.LogInfoMsg('\n\nStarting service on host %s port %d' % (opc_gate_host ,opc_gate_port))

        daemon = Pyro5.server.Daemon(host=opc_gate_host, port=opc_gate_port)
        daemon.register(opc(), "opc")

        socks = daemon.sockets
        while win32event.WaitForSingleObject(self.hWaitStop, 0) != win32event.WAIT_OBJECT_0:
            ins,outs,exs = select.select(socks,[],[],1)
            if ins:
                daemon.events(ins)

        daemon.shutdown()

def main():
    if len(sys.argv) == 1:
        try:
            evtsrc_dll = os.path.abspath(servicemanager.__file__)
            servicemanager.PrepareToHostSingle(OpcService)
            servicemanager.Initialize('zzzOpenOPCService', evtsrc_dll)
            servicemanager.StartServiceCtrlDispatcher()
        except win32service.error as details:
            if details.winerror == winerror.ERROR_FAILED_SERVICE_CONTROLLER_CONNECT:
                win32serviceutil.usage()
                print(' --foreground: Run OpenOPCService in foreground.')

    else:
        if sys.argv[1] == '--foreground':
            daemon = Pyro5.server.Daemon(host=opc_gate_host, port=opc_gate_port)
            daemon.register(opc(), 'opc')

            socks = set(daemon.sockets)
            while True:
                ins,outs,exs = select.select(socks,[],[],1)
                if ins:
                    daemon.events(ins)

            daemon.shutdown()
        else:
            win32serviceutil.HandleCommandLine(OpcService)

if __name__ == '__main__':
    main()



''' Supported platforms
    Linux x86_64
    Windows win32 and x86_64
    
    Requires setuptools >=43.0.0 for platform_system attribute
'''
import sys
from setuptools import setup

# OpenOPCService can only be installed on Windows systems
noservicepackage = len(sys.argv) > 2  and (sys.argv[2] == "--plat-name=manylinux1_x86_64" or sys.argv[2] == "--plat-name=win_amd64")

packagelist = ["OpenOPC"] if noservicepackage else ["OpenOPC", "OpenOPCService"]
entries= {'console_scripts': ['opc=OpenOPC.opc:main'] if noservicepackage else ['opc=OpenOPC.opc:main', 'opc-service=OpenOPCService.opcservice:main'] }
requires= ['Pyro5>=5.13.1'] if noservicepackage else ['Pyro5>=5.13.1', 'pywin32>=221; platform_system=="Windows"'] # pywin32>=221 allowed for Python 3.4 compatible version of pywin32

setup(packages=packagelist,
        install_requires=requires,
        entry_points=entries)



###########################################################################
#
# OpenOPC for Python OPC-DA Tools Library file
#
# A Windows only OPC-DA library file.
#
# Copyright (c) 2007-2012 Barry Barnreiter (barry_b@users.sourceforge.net)
# Copyright (c) 2014 Anton D. Kachalov (mouse@yandex.ru)
# Copyright (c) 2017 José A. Maita (jose.a.maita@gmail.com)
# Copyright (c) 2022 j3mg
#
###########################################################################
import re
import OpenOPC
import OpenOPC.systemhealth
import pythoncom
import pywintypes
import time
from OpenOPC.common import exceptional, get_error_str, quality_str, type_check, wild2regex, ACCESS_RIGHTS, BROWSER_TYPE, OPCError, OPC_QUALITY, OPC_STATUS

class ClientTools():
    def __init__(self, opc_class, opc_host):
        self.opc_class = opc_class
        self.opc_host = opc_host
        self.__open_serv__ = None 
        self.__open_host__ = None
        self.__open_port__ = None
        self.__open_guid__ = None
        self._prev_serv_time = None

    def set_opc_host(self, host):
        self.opc_host = host

    def set_gateway_settings(self, service, host, port, guid):
        self.__open_serv__ = service 
        self.__open_host__ = host
        self.__open_port__ = port
        self.__open_guid__ = guid

    def iproperties(self, _opc, tags, id=None):
        """Iterable version of properties()"""

        try:
            self._update_tx_time()
            pythoncom.CoInitialize()

            tags, single_tag, valid = type_check(tags)
            if not valid:
                raise TypeError("properties(): 'tags' parameter must be a string or a list of strings")

            try:
                id.remove(0)
                include_name = True
            except:
                include_name = False

            if id != None:
                descriptions= []

                if isinstance(id, list) or isinstance(id, tuple):
                    property_id = list(id)
                    single_property = False
                else:
                    property_id = [id]
                    single_property = True

                for i in property_id:
                    descriptions.append('Property id %d' % i)
            else:
                single_property = False

            properties = []

            for tag in tags:

                if id == None:
                    descriptions = []
                    property_id = []
                    count, property_id, descriptions, datatypes = _opc.QueryAvailableProperties(tag)

                    # Remove bogus negative property id (not sure why this sometimes happens)
                    tag_properties = list(map(lambda x, y: (x, y), property_id, descriptions))
                    property_id = [p for p, d in tag_properties if p > 0]
                    descriptions = [d for p, d in tag_properties if p > 0]

                property_id.insert(0, 0)
                values = []
                errors = []
                values, errors = _opc.GetItemProperties(tag, len(property_id)-1, property_id)

                property_id.pop(0)
                values = [str(v) if type(v) == pywintypes.TimeType else v for v in values]

                # Replace variant id with type strings
                try:
                    i = property_id.index(1)
                    values[i] = vt[values[i]]
                except:
                    pass

                # Replace quality bits with quality strings
                try:
                    i = property_id.index(3)
                    values[i] = quality_str(values[i])
                except:
                    pass

                # Replace access rights bits with strings
                try:
                    i = property_id.index(5)
                    values[i] = ACCESS_RIGHTS[values[i]]
                except:
                    pass

                if id != None:
                    if single_property:
                        if single_tag:
                            tag_properties = values
                        else:
                            tag_properties = [values]
                    else:
                        tag_properties = list(map(lambda x, y: (x, y), property_id, values))
                else:
                    tag_properties = list(map(lambda x, y, z: (x, y, z), property_id, descriptions, values))
                    tag_properties.insert(0, (0, 'Item ID (virtual property)', tag))

                if include_name:    tag_properties.insert(0, (0, tag))
                if not single_tag:  tag_properties = [tuple([tag] + list(p)) for p in tag_properties]

                for p in tag_properties: yield p

        except pythoncom.com_error as err:
            error_msg = 'properties: %s' % get_error_str(err, _opc)
            raise OPCError(error_msg)

    def properties(self, _opc, tags, id=None):
        """Return list of property tuples (id, name, value) for the specified tag(s) """

        if type(tags) not in (list, tuple) and type(id) not in (type(None), list, tuple):
            single = True
        else:
            single = False

        props = self.iproperties(_opc, tags, id)

        if single:
            return list(props)[0]
        else:
            return list(props)

    def ilist(self, _opc, paths='*', recursive=False, flat=False, include_type=False):
        """Iterable version of list()"""

        try:
            self._update_tx_time()
            pythoncom.CoInitialize()

            try:
                browser = _opc.CreateBrowser()
            # For OPC servers that don't support browsing
            except:
                return {}

            paths, single, valid = type_check(paths)
            if not valid:
                raise TypeError("list(): 'paths' parameter must be a string or a list of strings")

            if len(paths) == 0: paths = ['*']
            nodes = {}

            for path in paths:

                if flat:
                    browser.MoveToRoot()
                    browser.Filter = ''
                    browser.ShowLeafs(True)

                    pattern = re.compile('^%s$' % wild2regex(path) , re.IGNORECASE)
                    matches = filter(pattern.search, browser)
                    if include_type:  matches = [(x, node_type) for x in matches]

                    for node in matches: yield node
                    continue

                queue = []
                queue.append(path)

                while len(queue) > 0:
                    tag = queue.pop(0)

                    browser.MoveToRoot()
                    browser.Filter = ''
                    pattern = None

                    path_str = '/'
                    path_list = tag.replace('.','/').split('/')
                    path_list = [p for p in path_list if len(p) > 0]
                    found_filter = False
                    path_postfix = '/'

                    for i, p in enumerate(path_list):
                        if found_filter:
                            path_postfix += p + '/'
                        elif p.find('*') >= 0:
                            pattern = re.compile('^%s$' % wild2regex(p) , re.IGNORECASE)
                            found_filter = True
                        elif len(p) != 0:
                            pattern = re.compile('^.*$')
                            browser.ShowBranches()

                            # Branch node, so move down
                            if len(browser) > 0:
                                try:
                                    browser.MoveDown(p)
                                    path_str += p + '/'
                                except:
                                    if i < len(path_list)-1: return
                                    pattern = re.compile('^%s$' % wild2regex(p) , re.IGNORECASE)

                            # Leaf node, so append all remaining path parts together
                            # to form a single search expression
                            else:
                                p = string.join(path_list[i:], '.')
                                pattern = re.compile('^%s$' % wild2regex(p) , re.IGNORECASE)
                                break

                    browser.ShowBranches()

                    if len(browser) == 0:
                        browser.ShowLeafs(False)
                        lowest_level = True
                        node_type = 'Leaf'
                    else:
                        lowest_level = False
                        node_type = 'Branch'

                    matches = filter(pattern.search, browser)

                    if not lowest_level and recursive:
                        queue += [path_str + x + path_postfix for x in matches]
                    else:
                        if lowest_level:  matches = [exceptional(browser.GetItemID,x)(x) for x in matches]
                        if include_type:  matches = [(x, node_type) for x in matches]
                        for node in matches:
                            if not node in nodes: yield node
                            nodes[node] = True

        except pythoncom.com_error as err:
            error_msg = 'list: %s' % get_error_str(err, _opc)
            raise OPCError(error_msg)

    def list(self, _opc, paths='*', recursive=False, flat=False, include_type=False):
        """Return list of item nodes at specified path(s) (tree browser)"""

        nodes = self.ilist(_opc, paths, recursive, flat, include_type)
        return list(nodes)

    def servers(self, _opc, opc_host='localhost'):
        """Return list of available OPC servers"""

        try:
            pythoncom.CoInitialize()
            servers = _opc.GetOPCServers(opc_host)
            servers = [s for s in servers if s != None]
            return servers

        except pythoncom.com_error as err:
            error_msg = 'servers: %s' % get_error_str(err, _opc)
            raise OPCError(error_msg)

    def info(self, _opc):
        """Return list of (name, value) pairs about the OPC server"""

        try:
            self._update_tx_time()
            pythoncom.CoInitialize()

            info_list = []

            if self.__open_serv__:
                mode = 'OpenOPC'
            else:
                mode = 'DCOM'

            info_list += [('Protocol', mode)]

            if mode == 'OpenOPC':
                info_list += [('Gateway Host', '%s:%s' % (self.__open_host__, self.__open_port__))]
                info_list += [('Gateway Version', '%s' % OpenOPC.__version__)]
            info_list += [('Class', self.opc_class)]
            info_list += [('Client Name', _opc.ClientName)]
            info_list += [('OPC Host', self.opc_host)]
            info_list += [('OPC Server', _opc.ServerName)]
            info_list += [('State', OPC_STATUS[_opc.ServerState])]
            info_list += [('Version', '%d.%d (Build %d)' % (_opc.MajorVersion, _opc.MinorVersion, _opc.BuildNumber))]

            try:
                browser = _opc.CreateBrowser()
                browser_type = BROWSER_TYPE[browser.Organization]
            except:
                browser_type = 'Not Supported'

            info_list += [('Browser', browser_type)]
            info_list += [('Start Time', str(_opc.StartTime))]
            info_list += [('Current Time', str(_opc.CurrentTime))]
            info_list += [('Vendor', _opc.VendorInfo)]

            return info_list

        except pythoncom.com_error as err:
            error_msg = 'info: %s' % get_error_str(err, _opc)
            raise OPCError(error_msg)

    def ping(self, _opc):
        """Check if we are still talking to the OPC server"""
        try:
            pythoncom.CoInitialize()
            # Convert OPC server time to milliseconds
            timestr = str(_opc.CurrentTime)
            hours = timestr[11:12]
            minutes = timestr[14:15]
            seconds = timestr[17:]
            totalseconds = float(hours)*3600.0 + float(minutes)*60.0 + float(seconds[:9])
            opc_serv_time = int(totalseconds) * 1000000.0
            if opc_serv_time == self._prev_serv_time:
                return False
            else:
                self._prev_serv_time = opc_serv_time
                return True
        except pythoncom.com_error:
            return False

    def _update_tx_time(self):
        """Update the session's last transaction time in the Gateway Service"""
        if self.__open_serv__:
            self.__open_serv__._tx_times[self.__open_guid__] = time.time()
            return True
        return False;
